import bpy
import math
import mathutils

def make_edge_key(i1, i2):
	return (min(i1, i2), max(i1, i2))

class Edge:
	def __init__(self, me):
		if me.__class__==Edge:
			self._medge = me._medge
			self.vertices = me.vertices[:]
			self.smooth = me.smooth
		else:
			self._medge = me
			self.smooth = False
		self.key = me.key
		self.faces = []

	def __getattr__(self, attr):
		return getattr(self._medge, attr)

	def check_smooth(self, limit):
		if len(self.faces)!=2:
			return

		d = self.faces[0].normal.dot(self.faces[1].normal)
		self.smooth = ((d>limit and self.faces[0].use_smooth and self.faces[1].use_smooth) or d>0.99995)

	def other_face(self, f):
		if f.index==self.faces[0].index:
			if len(self.faces)>=2:
				return self.faces[1]
			else:
				return None
		else:
			return self.faces[0]

	def other_vertex(self, v):
		if v.index==self.vertices[0].index:
			return self.vertices[1]
		else:
			return self.vertices[0]


class Vertex:
	def __init__(self, mv):
		if mv.__class__==Vertex:
			self._mvert = mv._mvert
			self.uvs = mv.uvs[:]
			self.tan = mv.tan
			self.bino = mv.bino
		else:
			self._mvert = mv
			self.uvs = []
			self.tan = None
			self.bino = None
		self.index = mv.index
		self.co = mv.co
		self.normal = mv.normal
		self.flag = False
		self.edges = []
		self.faces = []
		self.groups = mv.groups[:]

	def __getattr__(self, attr):
		return getattr(self._mvert, attr)

	def __cmp__(self, other):
		if other is None:
			return 1
		return cmp(self.index, other.index)


class VertexGroup:
	def __init__(self, base):
		self._base = base
		self.group = base.group
		self.weight = base.weight

	def __getattr__(self, attr):
		return getattr(self._mvert, attr)


class Face:
	def __init__(self, mf):
		self._mface = mf
		self.index = mf.index
		self.edges = []
		self.vertices = mf.vertices[:]
		self.uvs = []
		self.flag = False

	def __getattr__(self, attr):
		return getattr(self._mface, attr)

	def __cmp__(self, other):
		if other is None:
			return 1
		return cmp(self.index, other.index)

	def pivot_vertex(self, v):
		n = self.vertices.index(v)
		return [(n+i)%len(self.vertices) for i in range(len(self.vertices))]

	def pivot_vertices(self, *vt):
		flags = [(v in vt) for v in self.vertices]
		l = len(self.vertices)
		for i in range(l):
			if flags[i] and not flags[(i+l-1)%l]:
				return self.vertices[i:]+self.vertices[:i]

	def get_edge(self, v1, v2):	
		key = make_edge_key(v1.index, v2.index)
		for e in self.edges:
			if e.key==key:
				return e
		raise KeyError("No edge %s"%(key,))

	def other_edge(self, e, v):
		for d in self.edges:
			if d!=e and v in d.vertices:
				return d

	def get_neighbors(self):
		neighbors = [e.other_face(self) for e in self.edges]
		return list(filter(bool, neighbors))


class Line:
	def __init__(self, e):
		self.edge = e
		self.vertices = e.vertices[:]
		self.flag = False


class UvLayer:
	def __init__(self, arg):
		if type(arg)==str:
			self._layer = None
			self.name = arg
		else:
			self._layer = arg
			self.name = arg.name
			self.uvs = [d.uv for d in self.data]

		self.unit = None
		self.hidden = False

		dot = self.name.find('.')
		if dot>=0:
			ext = self.name[dot:]
			if ext.startswith(".unit") and ext[5:].isdigit():
				self.unit = int(ext[5:])
			elif ext==".hidden":
				self.hidden = True

	def __getattr__(self, attr):
		return getattr(self._layer, attr)


class Mesh:
	def __init__(self, m):
		self._mesh = m

		self.winding_test = m.winding_test
		self.tbn_vecs = m.tbn_vecs
		self.vertex_groups = m.vertex_groups

		self.vertices = [Vertex(v) for v in self.vertices]
		self.faces = [Face(f) for f in self.polygons]
		self.edges = [Edge(e) for e in self.edges]
		self.edge_map = {e.key: e for e in self.edges}
		self.loops = self.loops[:]

		self.materials = self.materials[:]
		if self.use_uv=='NONE' or not self.uv_layers:
			self.uv_layers = []
		else:
			self.uv_layers = [UvLayer(u) for u in self.uv_layers]
			self.uv_layers = sorted([u for u in self.uv_layers if not u.hidden], key=(lambda u: (u.unit or 1000, u.name)))

			if self.use_uv=='UNIT0':
				self.uv_layers = [self.uv_layers[0]]

			next_unit = max((u.unit+1 for u in self.uv_layers if u.unit is not None), default=0)
			for u in self.uv_layers:
				if not u.unit:
					u.unit = next_unit
					next_unit += 1

		for v in self.vertices:
			v.groups = [VertexGroup(g) for g in v.groups]

		for f in self.faces:
			if len(f.vertices)>4:
				raise ValueError("Ngons are not supported")

			f.vertices = [self.vertices[i] for i in f.vertices]
			for v in f.vertices:
				v.faces.append(f)

			for k in f.edge_keys:
				e = self.edge_map[k]
				e.faces.append(f)
				f.edges.append(e)

		for e in self.edges:
			e.vertices = [self.vertices[i] for i in e.vertices]
			for v in e.vertices:
				v.edges.append(e)

		if self.use_lines:
			self.lines = [Line(e) for e in self.edges if not e.faces]
		else:
			self.lines = []

	def __getattr__(self, attr):
		return getattr(self._mesh, attr)

	def transform(self, matrix):
		for v in self.vertices:
			v.co = matrix*v.co

	def splice(self, other):
		if len(self.uv_layers)!=len(other.uv_layers):
			raise ValueError("Meshes have incompatible UV layers")
		for i, u in enumerate(self.uv_layers):
			if u.name!=other.uv_layers[i].name:
				raise ValueError("Meshes have incompatible UV layers")

		material_map = []
		for m in other.materials:
			if m in self.materials:
				material_map.append(self.materials.index(m))
			else:
				material_map.append(len(self.materials))
				self.materials.append(m)

		for i, u in enumerate(self.uv_layers):
			u.uvs += other.uv_layers[i].uvs

		offset = len(self.vertices)
		self.vertices += other.vertices
		for v in self.vertices[offset:]:
			v.index += offset

		loop_offset = len(self.loops)
		self.loops += other.loops

		offset = len(self.faces)
		self.faces += other.faces
		for f in self.faces[offset:]:
			f.index += offset
			f.loop_start += loop_offset
			f.loop_indices = range(f.loop_start, f.loop_start+f.loop_total)
			if other.materials:
				f.material_index = material_map[f.material_index]

		offset = len(self.edges)
		self.edges += other.edges
		for e in self.edges[offset:]:
			e.index += offset
			e.key = make_edge_key(e.vertices[0].index, e.vertices[1].index)
			self.edge_map[e.key] = e

		self.lines += other.lines

	def prepare_smoothing(self, progress):
		smooth_limit = -1
		if self.smoothing=='NONE':
			for f in self.faces:
				f.use_smooth = False

			smooth_limit = 1
		elif self.use_auto_smooth:
			smooth_limit = math.cos(self.auto_smooth_angle)

		for e in self.edges:
			e.check_smooth(smooth_limit)

		progress.push_task("Sharp edges", 0.0, 0.7)
		self.split_vertices(self.find_smooth_group, progress)

		if self.smoothing!='BLENDER':
			progress.set_task("Updating normals", 0.7, 1.0)
			self.compute_normals(progress)

		progress.pop_task()

	def prepare_vertex_groups(self, obj):
		for v in self.vertices:
			if v.groups:
				weight_sum = sum(g.weight for g in v.groups)
				v.groups = sorted(v.groups, key=(lambda g: g.weight), reverse=True)[:self.max_groups_per_vertex]
				weight_scale = weight_sum/sum(g.weight for g in v.groups)
				for g in v.groups:
					g.weight *= weight_scale

		if obj.parent and obj.parent.type=="ARMATURE":
			armature = obj.parent.data
			bone_indices = {b.name: i for i, b in enumerate(armature.bones)}
			group_index_map = {i: i for i in range(len(obj.vertex_groups))}
			for g in first_obj.vertex_groups:
				if g.name in bone_indices:
					group_index_map[g.index] = bone_indices[g.name]

			for v in self.vertices:
				for g in v.groups:
					g.group = group_index_map[g.group]

	def prepare_uv(self, obj, progress):
		if obj.material_tex and self.use_uv!='NONE':
			layer = UvLayer("material_tex")

			if self.use_uv=='UNIT0':
				self.uv_layers = [layer]
				layer.unit = 0
			else:
				self.uv_layers.append(layer)
				layer.unit = max((u.unit+1 for u in self.uv_layers if u.unit is not None), default=0)

			layer.uvs = [None]*len(self.loops)
			for f in self.faces:
				uv = mathutils.Vector(((f.material_index+0.5)/len(self.materials), 0.5))
				for i in f.loop_indices:
					layer.uvs[i] = uv

		array_uv_layers = [t.uv_layer for m in self.materials if m.array_atlas for t in m.texture_slots if t and t.texture_coords=='UV']
		array_uv_layers = [u for u in self.uv_layers if u.name in array_uv_layers]

		if array_uv_layers:
			for f in self.faces:
				layer = 0
				if f.material_index<len(self.materials):
					mat = self.materials[f.material_index]
					if mat and mat.array_atlas:
						layer = mat.array_layer

				for l in array_uv_layers:
					for i in f.loop_indices:
						l.uvs[i] = mathutils.Vector((*l.uvs[i], layer))

		for f in self.faces:
			for u in self.uv_layers:
				f.uvs.append([u.uvs[i] for i in f.loop_indices])

		prog_count = len(self.uv_layers)
		prog_step = 0

		tbn_layer_index = -1
		if self.tbn_vecs:
			uv_names = [u.name for u in self.uv_layers]
			if self.tbn_uvtex in uv_names:
				prog_count += 1
				tbn_layer_index = uv_names.index(self.tbn_uvtex)
				progress.push_task_slice("Computing TBN", 0, prog_count)
				self.compute_tbn(tbn_layer_index, progress)
				progress.set_task_slice(self.tbn_uvtex, 1, prog_count)
				self.split_vertices(self.find_uv_group, progress, tbn_layer_index)
				progress.pop_task()
				prog_step = 2

		for i, u in enumerate(self.uv_layers):
			if i==tbn_layer_index:
				continue

			progress.push_task_slice(u.name, prog_step, prog_count)
			self.split_vertices(self.find_uv_group, progress, i)
			progress.pop_task()
			prog_step += 1

		for v in self.vertices:
			if v.faces:
				f = v.faces[0]
				i = f.vertices.index(v)
				v.uvs = [u[i] for u in f.uvs]
			else:
				v.uvs = [(0.0, 0.0)]*len(self.uv_layers)

	def split_vertices(self, find_group_func, progress, *args):
		groups = []
		for i in range(len(self.vertices)):
			v = self.vertices[i]
			for f in v.faces:
				f.flag = False

			vg = []
			for f in v.faces:
				if not f.flag:
					vg.append(find_group_func(v, f, *args))

			groups.append(vg)

			progress.set_progress(i*0.5/len(self.vertices))

		for i in range(len(self.vertices)):
			for g in groups[i][1:]:
				v = Vertex(self.vertices[i])
				v.index = len(self.vertices)
				self.vertices.append(v)

				v_edges = []
				for e in self.vertices[i].edges:
					e_faces_in_g = [f for f in e.faces if f in g]
					if e_faces_in_g:
						boundary = len(e_faces_in_g)<len(e.faces)
						v_edges.append((e, boundary, e_faces_in_g))

				for e, boundary, e_faces_in_g in v_edges:
					if boundary:
						ne = Edge(e)
						ne.index = len(self.edges)
						self.edges.append(ne)

						for f in e_faces_in_g:
							e.faces.remove(f)
							f.edges[f.edges.index(e)] = ne
							ne.faces.append(f)
						e = ne
					else:
						del self.edge_map[e.key]
						self.vertices[i].edges.remove(e)
						v.edges.append(e)

					e.vertices[e.vertices.index(self.vertices[i])] = v

					e.key = make_edge_key(e.vertices[0].index, e.vertices[1].index)
					self.edge_map[e.key] = e

				for f in g:
					self.vertices[i].faces.remove(f)
					f.vertices[f.vertices.index(self.vertices[i])] = v
					v.faces.append(f)

			progress.set_progress(0.5+i*0.5/len(self.vertices))

	def find_smooth_group(self, vertex, face):
		face.flag = True

		edges = [e for e in face.edges if vertex in e.vertices]

		group = [face]
		for e in edges:
			f = face
			while e.smooth:
				f = e.other_face(f)
				if not f or f.flag:
					break

				f.flag = True
				group.append(f)
				e = f.other_edge(e, vertex)

		return group

	def find_uv_group(self, vertex, face, index):
		uv = face.uvs[index][face.vertices.index(vertex)]
		face.flag = True

		group = [face]
		for f in vertex.faces:
			if not f.flag and f.uvs[index][f.vertices.index(vertex)]==uv:
				f.flag = True
				group.append(f)

		return group

	def compute_normals(self, progress):
		for i, v in enumerate(self.vertices):
			v.normal = mathutils.Vector()
			for f in v.faces:
				fv = f.pivot_vertices(v)
				edge1 = fv[1].co-fv[0].co
				edge2 = fv[-1].co-fv[0].co
				if edge1.length and edge2.length:
					v.normal += f.normal*edge1.angle(edge2)

			if v.normal.length:
				v.normal.normalize()
			else:
				v.normal = mathutils.Vector((0, 0, 1))

			progress.set_progress(i/len(self.vertices))

	def compute_tbn(self, index, progress):
		layer_uvs = self.uv_layers[index].uvs

		for i, v in enumerate(self.vertices):
			v.tan = mathutils.Vector()
			v.bino = mathutils.Vector()
			for f in v.faces:
				vi = f.pivot_vertex(v)
				uv0 = layer_uvs[f.loop_indices[vi[0]]]
				uv1 = layer_uvs[f.loop_indices[vi[1]]]
				uv2 = layer_uvs[f.loop_indices[vi[-1]]]
				du1 = uv1[0]-uv0[0]
				du2 = uv2[0]-uv0[0]
				dv1 = uv1[1]-uv0[1]
				dv2 = uv2[1]-uv0[1]
				edge1 = f.vertices[vi[1]].co-f.vertices[vi[0]].co
				edge2 = f.vertices[vi[-1]].co-f.vertices[vi[0]].co
				div = (du1*dv2-du2*dv1)
				if div:
					mul = edge1.angle(edge2)/div
					v.tan += (edge1*dv2-edge2*dv1)*mul
					v.bino += (edge2*du1-edge1*du2)*mul

			if v.tan.length:
				v.tan.normalize()
			if v.bino.length:
				v.bino.normalize()

			progress.set_progress(i/len(self.vertices))

	def drop_references(self):
		for v in self.vertices:
			v._mvert = None
			for g in v.groups:
				g._base = None
		for e in self.edges:
			e._medge = None
		for f in self.faces:
			f._mface = None
		for u in self.uv_layers:
			u._layer = None
		self._mesh = None

	def create_strip(self, face, max_len):
		# Find an edge with another unused face next to it
		edge = None
		for e in face.edges:
			other = e.other_face(face)
			if other and not other.flag:
				edge = e
				break

		if not edge:
			return None

		# Add initial vertices so that we'll complete the edge on the first
		# iteration
		vertices = face.pivot_vertices(*edge.vertices)
		if len(vertices)==3:
			result = [vertices[-1], vertices[0]]
		else:
			result = [vertices[-2], vertices[-1]]

		while 1:
			face.flag = True

			vertices = face.pivot_vertices(*result[-2:])
			k = len(result)%2

			# Quads need special handling because the winding of every other
			# triangle in the strip is reversed
			if len(vertices)==4 and not k:
				result.append(vertices[3])
			result.append(vertices[2])
			if len(vertices)==4 and k:
				result.append(vertices[3])

			if len(result)>=max_len:
				break

			# Hop over the last edge
			edge = face.get_edge(*result[-2:])
			face = edge.other_face(face)
			if not face or face.flag:
				break

		return result

def create_mesh_from_object(context, obj, progress):
	if obj.type!="MESH":
		raise Exception("Object is not a mesh")

	progress.push_task("Preparing mesh", 0.0, 0.3)

	objs = [(obj, mathutils.Matrix())]
	i = 0
	while i<len(objs):
		o, m = objs[i]
		i += 1
		for c in o.children:
			if c.type=="MESH" and c.compound:
				objs.append((c, m*c.matrix_local))

	mesh = None
	bmeshes = []
	for o, m in objs:
		bmesh = o.to_mesh(context.scene, True, "PREVIEW")
		bmeshes.append(bmesh)

		# Object.to_mesh does not copy custom properties
		bmesh.winding_test = o.data.winding_test
		bmesh.smoothing = o.data.smoothing
		bmesh.use_lines = o.data.use_lines
		bmesh.vertex_groups = o.data.vertex_groups
		bmesh.max_groups_per_vertex = o.data.max_groups_per_vertex
		bmesh.use_uv = o.data.use_uv
		bmesh.tbn_vecs = o.data.tbn_vecs
		bmesh.tbn_uvtex = o.data.tbn_uvtex

		me = Mesh(bmesh)
		me.transform(m)

		if mesh:
			mesh.splice(me)
		else:
			mesh = me

	progress.set_task("Smoothing", 0.3, 0.6)
	mesh.prepare_smoothing(progress)
	progress.set_task("Vertex groups", 0.6, 0.7)
	mesh.prepare_vertex_groups(obj)
	progress.set_task("Preparing UVs", 0.7, 1.0)
	mesh.prepare_uv(obj, progress)

	mesh.drop_references()
	for m in bmeshes:
		bpy.data.meshes.remove(m)

	progress.pop_task()

	return mesh
